from typing import ClassVar, Optional

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .port import Port


class NetworkService(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "name"
    __primarylabel__: ClassVar[Optional[str]] = "NetworkService"

    name: str
    description: Optional[str] = None


#
# OUTGOING RELATIONSHIPS
#


class NetworkServiceRunsOnPort(OntolocyRelationship):
    source: NetworkService
    target: Port

    __relationshiptype__: ClassVar[str] = "NETWORK_SERVICE_RUNS_ON_PORT"



from datetime import datetime
from typing import ClassVar, Optional

from pydantic import HttpUrl, StringConstraints, field_validator
from typing_extensions import Annotated

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .mitreattacktechnique import MitreAttackTechnique


class MitreAttackTactic(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "stix_id"
    __primarylabel__: ClassVar[str] = "MitreAttackTactic"

    stix_id: str
    stix_type: str
    stix_created: datetime
    stix_modified: datetime
    stix_spec_version: str = "2.1"
    stix_revoked: Optional[bool] = False

    attack_id: Annotated[
        str, StringConstraints(to_upper=True, pattern=r"TA\d{4}")
    ]  # noqa: F722
    attack_spec_version: str
    attack_version: str
    attack_shortname: str

    ref_url: HttpUrl
    name: str
    description: str

    @field_validator("stix_revoked")
    @classmethod
    def set_false(cls, v):
        if v is None:
            return False
        else:
            return v


#
# OUTGOING RELATIONSHIPS
#


class MitreTacticIncludesTechnique(OntolocyRelationship):
    source: MitreAttackTactic
    target: MitreAttackTechnique

    __relationshiptype__: ClassVar[str] = "MITRE_TACTIC_INCLUDES_TECHNIQUE"



from datetime import datetime
from typing import ClassVar, Optional

from pydantic import HttpUrl, field_validator

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .intrusionset import IntrusionSet
from .mitreattacksoftware import MitreAttackSoftware
from .mitreattacktechnique import MitreAttackTechnique


class MitreAttackCampaign(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "stix_id"
    __primarylabel__: ClassVar[str] = "MitreAttackCampaign"

    stix_id: str
    stix_type: str
    stix_created: datetime
    stix_modified: datetime
    stix_spec_version: str = "2.1"
    stix_revoked: Optional[bool] = False
    stix_first_seen: datetime
    stix_last_seen: datetime

    attack_spec_version: str
    attack_version: str
    attack_first_seen_citation: str
    attack_last_seen_citation: str
    attack_id: str

    ref_url: HttpUrl
    name: str
    description: str

    @field_validator("stix_revoked")
    @classmethod
    def set_false(cls, v):
        if v is None:
            return False
        else:
            return v


#
# OUTGOING RELATIONSHIPS
#


class MitreCampaignUsesTechnique(OntolocyRelationship):
    source: MitreAttackCampaign
    target: MitreAttackTechnique

    __relationshiptype__: ClassVar[str] = "MITRE_CAMPAIGN_USES_TECHNIQUE"


class MitreCampaignUsesSoftware(OntolocyRelationship):
    source: MitreAttackCampaign
    target: MitreAttackSoftware

    __relationshiptype__: ClassVar[str] = "MITRE_CAMPAIGN_USES_SOFTWARE"


class MitreCampaignAttributedTo(OntolocyRelationship):
    source: MitreAttackCampaign
    target: IntrusionSet

    __relationshiptype__: ClassVar[str] = "MITRE_CAMPAIGN_ATTRIBUTED_TO_INTRUSION_SET"



from datetime import datetime
from typing import ClassVar, Optional

from pydantic import StringConstraints
from typing_extensions import Annotated

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .cpe import CPE
from .cwe import CWE


class CVE(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "cve_id"
    __primarylabel__: ClassVar[Optional[str]] = "CVE"

    cve_id: Annotated[
        str, StringConstraints(to_upper=True, pattern=r"(CVE|cve)-\d{4}-\d{4,8}")
    ]  # noqa: F722
    published_date: Optional[datetime] = None
    assigner: Optional[str] = None
    description: Optional[str] = None


#
# OUTGOING RELATIONSHIPS
#


class CVERelatesToCWE(OntolocyRelationship):
    source: CVE
    target: CWE

    __relationshiptype__: ClassVar[str] = "CVE_RELATES_TO_CWE"


class CVERelatesToCPE(OntolocyRelationship):
    source: CVE
    target: CPE

    __relationshiptype__: ClassVar[str] = "CVE_RELATES_TO_CPE"



from typing import ClassVar, Optional

from pydantic import HttpUrl

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .actortype import ActorType
from .country import Country
from .mitreattacksoftware import MitreAttackSoftware
from .mitreattacktechnique import MitreAttackTechnique
from .threatactor import ThreatActor


class IntrusionSet(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "IntrusionSet"

    name: str
    unique_id: str
    description: Optional[str] = None
    url_reference: Optional[HttpUrl] = None

    def __str__(self):
        return self.name


#
# OUTGOING RELATIONSHIPS
#


class IntrusionSetAttributedToNation(OntolocyRelationship):
    source: IntrusionSet
    target: Country
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "INTRUSION_SET_ATTRIBUTED_TO_NATION"


class IntrusionSetLinkedToIntrusionSet(OntolocyRelationship):
    source: IntrusionSet
    target: IntrusionSet
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "INTRUSION_SET_LINKED_TO_INTRUSION_SET"


class IntrusionSetLinkedToThreatActor(OntolocyRelationship):
    source: IntrusionSet
    target: ThreatActor
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "INTRUSION_SET_LINKED_TO_THREAT_ACTOR"


class IntrusionSetIsOfType(OntolocyRelationship):
    source: IntrusionSet
    target: ActorType
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "INTRUSION_SET_IS_OF_TYPE"


class IntrusionSetUsesTechnique(OntolocyRelationship):
    source: IntrusionSet
    target: MitreAttackTechnique

    __relationshiptype__: ClassVar[str] = "INTRUSION_SET_USES_TECHNIQUE"


class IntrusionSetUsesSoftware(OntolocyRelationship):
    source: IntrusionSet
    target: MitreAttackSoftware

    __relationshiptype__: ClassVar[str] = "INTRUSION_SET_USES_SOFTWARE"



from datetime import date
from typing import ClassVar, Optional

from pydantic import HttpUrl

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .campaign import Campaign
from .country import Country
from .cve import CVE
from .cyberharm import CyberHarm
from .intrusionset import IntrusionSet
from .ip import IPAddressNode
from .mitreattacksoftware import MitreAttackSoftware
from .mitreattacktechnique import MitreAttackTechnique
from .sector import Sector
from .threatactor import ThreatActor


class Report(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "url_reference"
    __primarylabel__: ClassVar[Optional[str]] = "Report"

    title: str
    author: str
    url_reference: HttpUrl
    published_date: date
    summary: Optional[str] = None


#
# OUTGOING RELATIONSHIPS
#


class ReportMentionsIntrusionSet(OntolocyRelationship):
    source: Report
    target: IntrusionSet

    __relationshiptype__: ClassVar[str] = "REPORT_MENTIONS_INTRUSION_SET"


class ReportMentionsTechnique(OntolocyRelationship):
    source: Report
    target: MitreAttackTechnique

    __relationshiptype__: ClassVar[str] = "REPORT_MENTIONS_TECHNIQUE"


class ReportMentionsCVE(OntolocyRelationship):
    source: Report
    target: CVE

    __relationshiptype__: ClassVar[str] = "REPORT_MENTIONS_CVE"


class ReportMentionsSector(OntolocyRelationship):
    source: Report
    target: Sector

    __relationshiptype__: ClassVar[str] = "REPORT_MENTIONS_SECTOR"


class ReportMentionsCountry(OntolocyRelationship):
    source: Report
    target: Country

    __relationshiptype__: ClassVar[str] = "REPORT_MENTIONS_COUNTRY"


class ReportMentionsIP(OntolocyRelationship):
    source: Report
    target: IPAddressNode

    __relationshiptype__: ClassVar[str] = "REPORT_MENTIONS_IP"


class ReportIdentifiesIntrusionSet(OntolocyRelationship):
    source: Report
    target: IntrusionSet

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_INTRUSION_SET"


class ReportIdentifiesThreatActor(OntolocyRelationship):
    source: Report
    target: ThreatActor

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_THREAT_ACTOR"


class ReportIdentifiesTechnique(OntolocyRelationship):
    source: Report
    target: MitreAttackTechnique

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_TECHNIQUE"


class ReportIdentifiesCVE(OntolocyRelationship):
    source: Report
    target: CVE

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_CVE"


class ReportIdentifiesVictimSector(OntolocyRelationship):
    source: Report
    target: Sector

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_VICTIM_SECTOR"


class ReportIdentifiesVictimCountry(OntolocyRelationship):
    source: Report
    target: Country

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_VICTIM_COUNTRY"


class ReportIdentifiesSponsorCountry(OntolocyRelationship):
    source: Report
    target: Country

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_SPONSOR_COUNTRY"


class ReportIdentifiesCyberHarm(OntolocyRelationship):
    source: Report
    target: CyberHarm

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_CYBER_HARM"


class ReportIdentifiesCampaign(OntolocyRelationship):
    source: Report
    target: Campaign

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_CAMPAIGN"


class ReportIdentifiesSoftware(OntolocyRelationship):
    source: Report
    target: MitreAttackSoftware

    __relationshiptype__: ClassVar[str] = "REPORT_IDENTIFIES_SOFTWARE"



from typing import ClassVar, Optional

from pydantic import HttpUrl

from ..node import OntolocyNode


class CyberHarm(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "CyberHarm"

    name: str
    unique_id: str
    harm_type: str
    url_reference: HttpUrl
    description: Optional[str]



from typing import ClassVar, Optional

from pydantic import IPvAnyNetwork

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship


class ASN(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "number"
    __primarylabel__: ClassVar[Optional[str]] = "ASN"

    number: int
    network_name: str
    description: str
    cidr: IPvAnyNetwork
    country_code: str
    registry: str


#
# OUTGOING RELATIONSHIPS
#


class ASNHasWhoIsRegisteredContact(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "ASN_HAS_REGISTERED_CONTACT"

    source: ASN
    target: "Organisation"


from .organisation import Organisation  # noqa: E402

ASNHasWhoIsRegisteredContact.model_rebuild()



from typing import ClassVar, Optional

from pydantic import ValidationInfo, field_validator

from ..node import OntolocyNode


class JarmHash(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "jarm"
    __primarylabel__: ClassVar[str] = "JarmHash"

    jarm: str
    cipher_and_tls_version: Optional[str] = None
    tls_extensions_hash: Optional[str] = None

    @field_validator("cipher_and_tls_version")
    def set_cipher_and_tls_version(cls, v, info: ValidationInfo):
        """The first 30 characters of a jarm hash represent
        the cipher and tls version negotiated.
        https://engineering.salesforce.com/easily-identify-malicious-servers-on-the-internet-with-jarm-e095edac525a/
        """
        values = info.data

        if v is None and "jarm" in values:
            return values["jarm"][:30]
        else:
            return v

    @field_validator("tls_extensions_hash")
    def set_tls_extensions_hash(cls, v, info: ValidationInfo):
        """The last 32 characters of a jarm hash are a truncated sha256 hash
        of the tls extensions sent by the server.
        https://engineering.salesforce.com/easily-identify-malicious-servers-on-the-internet-with-jarm-e095edac525a/
        """
        values = info.data
        if v is None and "jarm" in values:
            return values["jarm"][-32:]
        else:
            return v



from typing import ClassVar, Optional

from pydantic import StringConstraints, ValidationInfo, field_validator
from typing_extensions import Annotated

from ..node import OntolocyNode


class CPE(OntolocyNode):
    """Defines the CPE (common platform enumeration) node type

    We only look at the part, vendor and product.
    We don't create a whole new node for every possible version.

    We take a 2.3 or 2.2 string and convert it to a 2.3 string
        without version/update/edition information.
    """

    __primaryproperty__: ClassVar[str] = "cpe"
    __primarylabel__: ClassVar[Optional[str]] = "CPE"

    cpe: Annotated[
        str,
        StringConstraints(
            pattern=(
                r"(cpe:2\.3:[aho\*\-](:(((\?*|\*?)([a-zA-Z0-9\-\._]|"  # noqa: F722
                r"(\\[\\\*\?!#$$%&'\(\)\+,/:;<=>@\[\]\^`\{\|}~]))+(\?*|\*?))|[\*\-])){5}"
                r"(:(([a-zA-Z]{2,3}(-([a-zA-Z]{2}|[0-9]{3}))?)|[\*\-]))(:(((\?*|\*?)"
                r"([a-zA-Z0-9\-\._]|(\\[\\\*\?!#$$%&'\(\)\+,/:;<=>@\[\]\^`\{\|}~]))+(\?*|\*?))|[\*\-])){4})"
            )
        ),
    ]
    cpe_version: str = "2.3"
    part: Optional[str] = None
    vendor: Optional[str] = None
    product: Optional[str] = None

    @field_validator("cpe", mode="before")
    def set_cpe(cls, v):
        # hack to handle CPEs with colons in (which will be escaped with a backslash)
        v = v.replace(r"\:", r"\;")

        cpe_parts = v.split(":")

        if cpe_parts[0] != "cpe":
            raise ValueError("Doesn't look like a valid CPE")

        if cpe_parts[1] == "2.3":
            v = f"cpe:{cpe_parts[1]}:{cpe_parts[2]}:{cpe_parts[3]}:{cpe_parts[4]}:*:*:*:*:*:*:*:*"

        # here's where we attempt to convert 2.2 format to 2.3
        elif cpe_parts[1] == "/a":
            v = f"cpe:2.3:a:{cpe_parts[2]}:{cpe_parts[3]}:*:*:*:*:*:*:*:*"

        elif cpe_parts[1] == "/o":
            v = f"cpe:2.3:o:{cpe_parts[2]}:{cpe_parts[3]}:*:*:*:*:*:*:*:*"

        elif cpe_parts[1] == "/h":
            v = f"cpe:2.3:h:{cpe_parts[2]}:{cpe_parts[3]}:*:*:*:*:*:*:*:*"

        else:
            raise ValueError("Doesn't look like a valid CPE")

        v = v.replace(r"\;", r"\:")

        return v

    @field_validator("part")
    def set_part(cls, v, info: ValidationInfo):
        values = info.data
        if v is None and "cpe" in values:
            cpe_parts = values["cpe"].split(":")
            if cpe_parts[2] in ["a", "h", "o"]:
                return cpe_parts[2]
            else:
                raise ValueError("Doesn't look like a valid CPE part (a/h/o)")
        else:
            return v

    @field_validator("vendor")
    def set_vendor(cls, v, info: ValidationInfo):
        values = info.data
        if v is None and "cpe" in values:
            cpe_parts = values["cpe"].split(":")
            return cpe_parts[3]
        else:
            return v

    @field_validator("product")
    def set_product(cls, v, info: ValidationInfo):
        values = info.data
        if v is None and "cpe" in values:
            cpe_parts = values["cpe"].split(":")
            return cpe_parts[4]
        else:
            return v



from datetime import date
from typing import ClassVar, Optional

from pydantic import AnyHttpUrl

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship

# from .cve import CVE

# from .report import Report


class Organisation(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "name"
    __primarylabel__: ClassVar[str] = "Organisation"

    name: str
    description: Optional[str] = None
    address: Optional[str] = None


#
# OUTGOING RELATIONSHIPS
#


class OrganisationAssignedCVSSToCVE(OntolocyRelationship):
    source: Organisation
    target: "CVE"

    version: str
    vector_string: str
    attack_vector: str
    attack_complexity: str
    privileges_required: str
    user_interaction: str
    scope: str
    confidentiality_impact: str
    integrity_impact: str
    availability_impact: str
    base_score: float
    base_severity: str
    exploitability_score: float
    impact_score: float

    __relationshiptype__: ClassVar[str] = "ORGANISATION_ASSIGNED_CVSS_TO_CVE"


class OrganisationReportedExploitationOfCVE(OntolocyRelationship):
    source: Organisation
    target: "CVE"

    reported_date: date
    url_reference: Optional[AnyHttpUrl]
    description: Optional[str]
    required_action: Optional[str]

    __relationshiptype__: ClassVar[str] = "ORGANISATION_REPORTED_EXPLOITATION_OF_CVE"


class OrganisationPublishedThreatReport(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "ORGANISATION_PUBLISHED_THREAT_REPORT"

    source: Organisation
    target: "Report"

    context: Optional[str]
    url_reference: Optional[AnyHttpUrl]


from .cve import CVE  # noqa: E402
from .report import Report  # noqa: E402

OrganisationAssignedCVSSToCVE.model_rebuild()
OrganisationPublishedThreatReport.model_rebuild()
OrganisationReportedExploitationOfCVE.model_rebuild()



from typing import ClassVar

from ..node import OntolocyNode


class CobaltStrikeWatermark(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "value"
    __primarylabel__: ClassVar[str] = "CobaltStrikeWatermark"

    value: int



from typing import ClassVar

from ontolocy import OntolocyNode, DataOrigin


def test_define_node():
    class CyberPerson(OntolocyNode):
        __primaryproperty__: ClassVar[str] = "name"
        __primarylabel__: ClassVar[str] = "CyberPerson"

        name: str

    person = CyberPerson(name="Foo")

    assert person.name == "Foo"


def test_ingest_node(use_graph):
    class CyberPerson(OntolocyNode):
        __primaryproperty__: ClassVar[str] = "name"
        __primarylabel__: ClassVar[str] = "CyberPerson"

        name: str

    person = CyberPerson(name="Foo")

    data_origin = DataOrigin(name="Test Case")

    person.ingest(data_origin=data_origin)

    cypher = """
    MATCH (o:DataOrigin)-[:ORIGIN_GENERATED]->(cp:CyberPerson)
    WHERE o.name = 'Test Case' AND cp.name = 'Foo'
    RETURN COUNT(DISTINCT cp)
    """

    result = use_graph.evaluate_query_single(cypher)

    assert result == 1



import hashlib
from typing import ClassVar, Optional

from pydantic import ValidationInfo, field_validator

from ..node import OntolocyNode


class Banner(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "sha1"
    __primarylabel__: ClassVar[str] = "Banner"

    banner: str
    sha1: Optional[str] = None

    @field_validator("sha1")
    def set_sha1(cls, v, info: ValidationInfo):
        values = info.data
        if v is None and "banner" in values:
            return hashlib.sha1(values["banner"].encode()).hexdigest()
        else:
            return v



from typing import ClassVar, Optional
from uuid import UUID, uuid4

from pydantic import ValidationInfo, field_validator

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from ..utils import generate_deterministic_uuid
from .host import Host


class UserAccount(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "UserAccount"

    username: str
    local_hostname: Optional[str] = None
    namespace: Optional[str] = None
    unique_id: Optional[UUID] = None

    def __str__(self) -> str:
        return self.username

    @field_validator("namespace")
    def set_namespace(cls, v) -> str:
        if v is None:
            return str(uuid4())

        else:
            return v

    @field_validator("unique_id")
    def generate_instance_id(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data

        if v is None:
            key_values = [
                values["username"],
                values["local_hostname"],
                values["namespace"],
            ]

            v = generate_deterministic_uuid(key_values)

        return v


class UserAccountAuthorizedOnHost(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "USER_ACCOUNT_AUTHORIZED_ON_HOST"

    source: UserAccount
    target: Host



from datetime import datetime
from typing import ClassVar, Optional

from pydantic import HttpUrl

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .country import Country
from .cve import CVE
from .cyberharm import CyberHarm
from .intrusionset import IntrusionSet
from .mitreattacktechnique import MitreAttackTechnique
from .sector import Sector


class Campaign(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "Campaign"

    title: str
    url_reference: HttpUrl
    activity_datetime: datetime
    summary: Optional[str]
    unique_id: str

    def __str__(self) -> str:
        return self.title


#
# OUTGOING RELATIONSHIPS
#


class CampaignByIntrusionSet(OntolocyRelationship):
    source: Campaign
    target: IntrusionSet
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "CAMPAIGN_BY_INTRUSION_SET"


class CampaignUsesTechnique(OntolocyRelationship):
    source: Campaign
    target: MitreAttackTechnique
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "CAMPAIGN_USES_TECHNIQUE"


class CampaignUsesCVE(OntolocyRelationship):
    source: Campaign
    target: CVE
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "CAMPAIGN_USES_CVE"


class CampaignTargetsSector(OntolocyRelationship):
    source: Campaign
    target: Sector
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "CAMPAIGN_TARGETS_SECTOR"


class CampaignTargetsCountry(OntolocyRelationship):
    source: Campaign
    target: Country
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "CAMPAIGN_TARGETS_COUNTRY"


class CampaignCausedCyberHarm(OntolocyRelationship):
    source: Campaign
    target: CyberHarm
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "CAMPAIGN_CAUSED_CYBER_HARM"



from datetime import datetime
from typing import ClassVar, Optional
from uuid import UUID

from pydantic import ValidationInfo, field_validator

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from ..utils import generate_deterministic_uuid


class DNSRecord(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "DNSRecord"

    type: str
    name: str
    content: str

    unique_id: Optional[UUID] = None

    @field_validator("unique_id")
    def generate_dnsrecord_uuid(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data

        if v is None:
            key_values = [
                values["type"],
                values["name"],
                values["content"],
            ]

            v = generate_deterministic_uuid(key_values)

        return v


class DNSRecordPointsToIPAddress(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "DNS_RECORD_POINTS_TO_IP_ADDRESS"

    source: DNSRecord
    target: "IPAddressNode"

    observation_date: datetime


class DNSRecordPointsToDomainName(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "DNS_RECORD_POINTS_TO_DOMAIN_NAME"

    source: DNSRecord
    target: "DomainName"

    observation_date: datetime


class DNSRecordForDomain(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "DNS_RECORD_FOR_DOMAIN"

    source: DNSRecord
    target: "DomainName"

    observation_date: datetime


from .domainname import DomainName  # noqa: E402

DNSRecordPointsToDomainName.model_rebuild()
DNSRecordForDomain.model_rebuild()

from .ip import IPAddressNode  # noqa: E402

DNSRecordPointsToIPAddress.model_rebuild()



from typing import ClassVar, Optional

from ..node import OntolocyNode


class X509Certificate(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "sha1"
    __primarylabel__: ClassVar[str] = "X509Certificate"

    sha1: str
    sha256: Optional[str] = None
    md5: Optional[str] = None

    serial_number: Optional[str] = None

    issuer_country: Optional[str] = None
    issuer_cn: Optional[str] = None
    issuer_organisation: Optional[str] = None
    issuer_locality: Optional[str] = None
    issuer_state: Optional[str] = None
    issuer_ou: Optional[str] = None

    subject_country: Optional[str] = None
    subject_cn: Optional[str] = None
    subject_organisation: Optional[str] = None
    subject_locality: Optional[str] = None
    subject_state: Optional[str] = None
    subject_ou: Optional[str] = None



from typing import ClassVar, Optional

from pydantic import StringConstraints, field_validator
from typing_extensions import Annotated

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .host import Host


class MACAddress(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "mac_address"
    __primarylabel__: ClassVar[Optional[str]] = "MACAddress"

    mac_address: Annotated[
        str,
        StringConstraints(
            pattern=r"^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$"  # noqa: F722
        ),
    ]

    @field_validator("mac_address")
    def format_mac_address(cls, v: str) -> str:
        """We want to standardize on colons and uppercase letters"""

        v = v.upper()
        v = v.replace("-", ":")

        return v


class MACAddressAssignedToHost(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "MAC_ADDRESS_ASSIGNED_TO_HOST"

    source: MACAddress
    target: Host

    interface: Optional[str] = None



from datetime import datetime
from typing import ClassVar, Optional

from pydantic import HttpUrl, StringConstraints, field_validator
from typing_extensions import Annotated

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship


class MitreAttackTechnique(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "stix_id"
    __primarylabel__: ClassVar[str] = "MitreAttackTechnique"

    stix_id: str
    stix_type: str
    stix_created: datetime
    stix_modified: datetime
    stix_spec_version: str = "2.1"
    stix_revoked: Optional[bool] = False

    attack_id: Annotated[
        str, StringConstraints(to_upper=True, pattern=r"T\d{4}(?:\.\d{3})?")
    ]  # noqa: F722
    attack_spec_version: str
    attack_subtechnique: Optional[bool] = False
    attack_version: str

    ref_url: HttpUrl
    name: str
    description: str

    @field_validator("stix_revoked", "attack_subtechnique")
    @classmethod
    def set_false(cls, v):
        if v is None:
            return False
        else:
            return v

    def __str__(self) -> str:
        return f"{self.attack_id}: {self.name}"


#
# OUTGOING RELATIONSHIPS
#


class MitreSubtechniqueOf(OntolocyRelationship):
    source: MitreAttackTechnique
    target: MitreAttackTechnique

    __relationshiptype__: ClassVar[str] = "MITRE_SUBTECHNIQUE_OF"



from abc import ABC
from typing import Optional

import pandas as pd
from neontology import BaseRelationship

from .dataorigin import DataOrigin


class OntolocyRelationship(BaseRelationship, ABC, validate_default=True):
    data_origin_name: Optional[str] = None
    data_origin_reference: Optional[str] = None
    data_origin_license: Optional[str] = None
    data_origin_sharing: Optional[str] = None

    @classmethod
    def ingest_df(
        cls,
        df: pd.DataFrame,
        data_origin: Optional[DataOrigin] = None,
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ):
        input_df = df.copy()

        if data_origin is not None:
            input_df["data_origin_name"] = data_origin.name
            input_df["data_origin_reference"] = data_origin.reference
            input_df["data_origin_license"] = data_origin.license
            input_df["data_origin_sharing"] = data_origin.sharing

        return cls.merge_df(input_df, source_prop=source_prop, target_prop=target_prop)



from typing import ClassVar, Optional
from uuid import UUID

from pydantic import ValidationInfo, field_validator

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from ..utils import generate_deterministic_uuid
from .cobaltstrikewatermark import CobaltStrikeWatermark
from .listeningsocket import ListeningSocket


class CobaltStrikeBeacon(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "CobaltStrikeBeacon"

    beacontype: int
    port: int
    sleeptime: int
    maxgetsize: int
    jitter: int
    c2server: str
    useragent: str
    submituri: str
    watermark: Optional[int] = None

    unique_id: Optional[UUID] = None

    @field_validator("unique_id")
    def generate_dnsrecord_uuid(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data
        if v is None:
            key_values = [
                values["beacontype"],
                values["port"],
                values["c2server"],
                values["watermark"],
            ]

            v = generate_deterministic_uuid(key_values)

        return v


class CobaltStikeBeaconCollectedFrom(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "COBALT_STRIKE_BEACON_COLLECTED_FROM"

    source: CobaltStrikeBeacon
    target: ListeningSocket


class CobaltStikeBeaconHasWatermark(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "COBALT_STRIKE_BEACON_HAS_WATERMARK"

    source: CobaltStrikeBeacon
    target: CobaltStrikeWatermark



from typing import ClassVar, Optional

from pydantic import HttpUrl

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .actortype import ActorType
from .country import Country


class ThreatActor(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "ThreatActor"

    name: str
    unique_id: str
    description: Optional[str]
    url_reference: Optional[HttpUrl]

    def __str__(self):
        return self.name


#
# OUTGOING RELATIONSHIPS
#


class ThreatActorAttributedToNation(OntolocyRelationship):
    source: ThreatActor
    target: Country
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "THREAT_ACTOR_ATTRIBUTED_TO_NATION"


class ThreatActorIsOfType(OntolocyRelationship):
    source: ThreatActor
    target: ActorType
    url_reference: Optional[HttpUrl]

    __relationshiptype__: ClassVar[str] = "THREAT_ACTOR_IS_OF_TYPE"


class ThreatActorLinkedToThreatActor(OntolocyRelationship):
    source: ThreatActor
    target: ThreatActor
    url_reference: Optional[HttpUrl]
    context: Optional[str]

    __relationshiptype__: ClassVar[str] = "THREAT_ACTOR_LINKED_TO_THREAT_ACTOR"



import re
from datetime import datetime
from typing import ClassVar, Optional

from pydantic import field_validator

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship


class DomainName(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "name"
    __primarylabel__: ClassVar[Optional[str]] = "DomainName"

    name: str

    @field_validator("name")
    @classmethod
    def validate_domain(cls, v: str) -> str:
        pattern = r"((?=[a-z0-9-]{1,63}\.)(xn--)?[a-z0-9]+(-[a-z0-9]+)*\.)+[a-z]{2,63}"
        m = re.match(pattern, v)

        if m is None:
            raise ValueError("Doesn't look like a valid domain")

        return v


class DomainNameHasDNSRecord(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "DOMAIN_NAME_HAS_DNS_RECORD"

    source: DomainName
    target: "DNSRecord"

    observation_date: datetime


from .dnsrecord import DNSRecord  # noqa: E402

DomainNameHasDNSRecord.model_rebuild()



from typing import ClassVar

from pydantic import AnyUrl

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship


class URLNode(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "url"
    __primarylabel__: ClassVar[str] = "URL"

    url: AnyUrl


class UrlRedirectsToUrl(OntolocyRelationship):

    __relationshiptype__: ClassVar[str] = "URL_REDIRECTS_TO_URL"

    source: URLNode
    target: URLNode



from enum import Enum
from ipaddress import ip_address
from typing import ClassVar, Optional
from uuid import UUID, uuid4

from pydantic import IPvAnyAddress, ValidationInfo, field_validator

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from ..utils import generate_deterministic_uuid
from .asn import ASN
from .country import Country
from .cpe import CPE
from .listeningsocket import ListeningSocket
from .macaddress import MACAddress


class IPVersionEnum(str, Enum):
    ipv4 = "ipv4"
    ipv6 = "ipv6"


class IPAddressNode(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[str] = "IPAddress"

    ip_address: IPvAnyAddress
    ip_version: Optional[IPVersionEnum] = None
    private: Optional[bool] = None
    namespace: Optional[str] = None

    unique_id: Optional[UUID] = None

    def __str__(self) -> str:
        return str(self.ip_address)

    @field_validator("ip_address", mode="before")
    @classmethod
    def refang_ip(cls, v: str):
        """Some reports will 'de-fang' an IP with square brackets.
        Remove them for consistency"""
        return v.replace("[", "").replace("]", "")

    @field_validator("ip_version")
    def mark_ip_version(cls, v, info: ValidationInfo):
        values = info.data
        versions = {4: IPVersionEnum.ipv4, 6: IPVersionEnum.ipv6}
        if v is None and "ip_address" in values:
            return versions[ip_address(values["ip_address"]).version]
        else:
            return v

    @field_validator("private")
    def mark_private(cls, v, info: ValidationInfo):
        values = info.data
        if v is None and "ip_address" in values:
            return ip_address(values["ip_address"]).is_private
        else:
            return v

    @field_validator("namespace")
    def set_namespace(cls, v, info: ValidationInfo):
        values = info.data
        if values["private"] is False:
            return None

        elif v is None:
            return str(uuid4())

        else:
            return v

    @field_validator("unique_id")
    def generate_instance_id(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data
        if v is None:
            key_values = [values["ip_address"], values["namespace"]]

            v = generate_deterministic_uuid(key_values)

        return v


class IPAddressHasOpenPort(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "IP_ADDRESS_HAS_OPEN_PORT"

    source: IPAddressNode
    target: ListeningSocket


class IPAddressBelongsToASN(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "IP_ADDRESS_BELONGS_TO_ASN"

    source: IPAddressNode
    target: ASN


class IPAddressLocatedInCountry(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "IP_ADDRESS_LOCATED_IN_COUNTRY"

    source: IPAddressNode
    target: Country


class IPAddressIdentifiedAsPlatform(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "IP_ADDRESS_IDENTIFIED_AS_PLATFORM"

    source: IPAddressNode
    target: CPE


class IPAddressMapsToMACAddress(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "IP_ADDRESS_MAPS_TO_MAC_ADDRESS"

    source: IPAddressNode
    target: MACAddress


class IPAddressObservedWithHostname(OntolocyRelationship):
    __relationshiptype__: ClassVar[str] = "IP_ADDRESS_OBSERVED_WITH_HOSTNAME"

    source: IPAddressNode
    target: "DomainName"


from .domainname import DomainName  # noqa: E402

IPAddressObservedWithHostname.model_rebuild()



from enum import Enum
from typing import ClassVar, Optional
from uuid import UUID, uuid4

from pydantic import ValidationInfo, field_validator

from ..node import OntolocyNode
from ..utils import generate_deterministic_uuid


class HostOSEnum(str, Enum):
    linux = "linux"
    windows = "windows"
    macos = "macos"


class Host(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "Host"

    hostname: str
    os: Optional[HostOSEnum] = None
    namespace: Optional[str] = None
    unique_id: Optional[UUID] = None

    def __str__(self) -> str:
        return self.hostname

    @field_validator("namespace")
    def set_namespace(cls, v):
        if v is None:
            return str(uuid4())

        else:
            return v

    @field_validator("unique_id")
    def generate_instance_id(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data

        if v is None:
            key_values = [values["hostname"], values["namespace"]]

            v = generate_deterministic_uuid(key_values)

        return v



from typing import ClassVar, Optional
from uuid import UUID

from pydantic import AnyHttpUrl, ValidationInfo, field_validator

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from ..utils import generate_deterministic_uuid
from .mitreattacktechnique import MitreAttackTechnique


class Control(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "Control"

    control_id: str
    name: str
    framework: str
    version: Optional[str] = None
    framework_version: Optional[str] = None
    description: Optional[str] = None
    context: Optional[str] = None
    unique_id: Optional[str] = None
    url_reference: Optional[AnyHttpUrl] = None

    def __str__(self) -> str:
        return f"{self.control_id}: {self.name}"

    @field_validator("unique_id")
    def generate_uuid(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data
        if v is None:
            key_values = [
                values["control_id"],
                values["version"],
                values["framework"],
                values["framework_version"],
            ]

            # the unique id should be a string rather than a UUID
            v = str(generate_deterministic_uuid(key_values))

        return v


#
# OUTGOING RELATIONSHIPS
#


class ControlRelatedToControl(OntolocyRelationship):
    source: Control
    target: Control

    context: Optional[str]
    url_reference: Optional[AnyHttpUrl]

    __relationshiptype__: ClassVar[str] = "CONTROL_RELATED_TO_CONTROL"


class ControlHasParentControl(OntolocyRelationship):
    source: Control
    target: Control

    context: Optional[str]
    url_reference: Optional[AnyHttpUrl]

    __relationshiptype__: ClassVar[str] = "CONTROL_HAS_PARENT_CONTROL"


class ControlMitigatesAttackTechnique(OntolocyRelationship):
    source: Control
    target: MitreAttackTechnique

    url_reference: Optional[AnyHttpUrl]

    __relationshiptype__: ClassVar[str] = "CONTROL_MITIGATES_ATTACK_TECHNIQUE"



from ontolocy.dataorigin import DataOrigin, OriginGenerated
from ontolocy.models.actortype import ActorType
from ontolocy.models.asn import ASN, ASNHasWhoIsRegisteredContact
from ontolocy.models.banner import Banner
from ontolocy.models.campaign import (
    Campaign,
    CampaignByIntrusionSet,
    CampaignCausedCyberHarm,
    CampaignTargetsCountry,
    CampaignTargetsSector,
    CampaignUsesCVE,
    CampaignUsesTechnique,
)
from ontolocy.models.capecpattern import (
    CAPECPattern,
    CAPECPatternMapsToAttackTechnique,
    CAPECPatternRelatesToCWE,
)
from ontolocy.models.cobaltstrikebeacon import (
    CobaltStikeBeaconCollectedFrom,
    CobaltStikeBeaconHasWatermark,
    CobaltStrikeBeacon,
)
from ontolocy.models.cobaltstrikewatermark import CobaltStrikeWatermark
from ontolocy.models.control import (
    Control,
    ControlHasParentControl,
    ControlMitigatesAttackTechnique,
    ControlRelatedToControl,
)
from ontolocy.models.country import Country
from ontolocy.models.cpe import CPE
from ontolocy.models.cve import CVE, CVERelatesToCPE, CVERelatesToCWE
from ontolocy.models.cwe import CWE
from ontolocy.models.cyberharm import CyberHarm
from ontolocy.models.dnsrecord import (
    DNSRecord,
    DNSRecordForDomain,
    DNSRecordPointsToDomainName,
    DNSRecordPointsToIPAddress,
)
from ontolocy.models.domainname import DomainName, DomainNameHasDNSRecord
from ontolocy.models.exploit import Exploit, ExploitExploitsVulnerability
from ontolocy.models.host import Host
from ontolocy.models.intrusionset import (
    IntrusionSet,
    IntrusionSetAttributedToNation,
    IntrusionSetIsOfType,
    IntrusionSetLinkedToIntrusionSet,
    IntrusionSetLinkedToThreatActor,
    IntrusionSetUsesSoftware,
    IntrusionSetUsesTechnique,
)
from ontolocy.models.ip import (
    IPAddressBelongsToASN,
    IPAddressHasOpenPort,
    IPAddressIdentifiedAsPlatform,
    IPAddressLocatedInCountry,
    IPAddressMapsToMACAddress,
    IPAddressNode,
    IPAddressObservedWithHostname,
)
from ontolocy.models.jarmhash import JarmHash
from ontolocy.models.listeningsocket import (
    ListeningSocket,
    ListeningSocketUsesPort,
    OpenPortHasJarmHash,
    OpenPortPresentsBanner,
    OpenPortPresentsX509Certificate,
    ServiceHostsURL,
    ServiceIdentifiedAsPlatform,
)
from ontolocy.models.macaddress import MACAddress, MACAddressAssignedToHost
from ontolocy.models.mitreattackcampaign import (
    MitreAttackCampaign,
    MitreCampaignAttributedTo,
    MitreCampaignUsesSoftware,
    MitreCampaignUsesTechnique,
)
from ontolocy.models.mitreattacksoftware import (
    MitreAttackSoftware,
    MitreSoftwareUsesTechnique,
)
from ontolocy.models.mitreattacktactic import (
    MitreAttackTactic,
    MitreTacticIncludesTechnique,
)
from ontolocy.models.mitreattacktechnique import (
    MitreAttackTechnique,
    MitreSubtechniqueOf,
)
from ontolocy.models.networkservice import NetworkService, NetworkServiceRunsOnPort
from ontolocy.models.organisation import (
    Organisation,
    OrganisationAssignedCVSSToCVE,
    OrganisationPublishedThreatReport,
    OrganisationReportedExploitationOfCVE,
)
from ontolocy.models.port import Port
from ontolocy.models.report import (
    Report,
    ReportIdentifiesCampaign,
    ReportIdentifiesCVE,
    ReportIdentifiesCyberHarm,
    ReportIdentifiesIntrusionSet,
    ReportIdentifiesSoftware,
    ReportIdentifiesSponsorCountry,
    ReportIdentifiesTechnique,
    ReportIdentifiesVictimCountry,
    ReportIdentifiesVictimSector,
    ReportMentionsCountry,
    ReportMentionsCVE,
    ReportMentionsIntrusionSet,
    ReportMentionsIP,
    ReportMentionsSector,
    ReportMentionsTechnique,
)
from ontolocy.models.sector import Sector
from ontolocy.models.threatactor import (
    ThreatActor,
    ThreatActorAttributedToNation,
    ThreatActorIsOfType,
)
from ontolocy.models.url import URLNode, UrlRedirectsToUrl
from ontolocy.models.useraccount import UserAccount, UserAccountAuthorizedOnHost
from ontolocy.models.x509certificate import X509Certificate
from ontolocy.node import OntolocyNode
from ontolocy.relationship import OntolocyRelationship
from ontolocy.utils import init_ontolocy

__all__ = [
    "DataOrigin",
    "OriginGenerated",
    "OntolocyNode",
    "OntolocyRelationship",
    "init_ontolocy",
    # models
    "ActorType",
    "ASN",
    "ASNHasWhoIsRegisteredContact",
    "Banner",
    "Campaign",
    "CampaignByIntrusionSet",
    "CampaignUsesTechnique",
    "CampaignUsesCVE",
    "CampaignTargetsSector",
    "CampaignTargetsCountry",
    "CampaignCausedCyberHarm",
    "CAPECPattern",
    "CAPECPatternMapsToAttackTechnique",
    "CAPECPatternRelatesToCWE",
    "CobaltStrikeBeacon",
    "CobaltStikeBeaconCollectedFrom",
    "CobaltStikeBeaconHasWatermark",
    "CobaltStrikeWatermark",
    "Control",
    "ControlHasParentControl",
    "ControlMitigatesAttackTechnique",
    "ControlRelatedToControl",
    "Country",
    "CPE",
    "CVE",
    "CVERelatesToCPE",
    "CVERelatesToCWE",
    "CWE",
    "CyberHarm",
    "DNSRecord",
    "DNSRecordForDomain",
    "DNSRecordPointsToDomainName",
    "DNSRecordPointsToIPAddress",
    "DomainName",
    "DomainNameHasDNSRecord",
    "Exploit",
    "ExploitExploitsVulnerability",
    "Host",
    "IntrusionSet",
    "IntrusionSetAttributedToNation",
    "IntrusionSetLinkedToIntrusionSet",
    "IntrusionSetIsOfType",
    "IntrusionSetLinkedToThreatActor",
    "IntrusionSetUsesSoftware",
    "IntrusionSetUsesTechnique",
    "IPAddressNode",
    "IPAddressBelongsToASN",
    "IPAddressHasOpenPort",
    "IPAddressIdentifiedAsPlatform",
    "IPAddressLocatedInCountry",
    "IPAddressMapsToMACAddress",
    "IPAddressObservedWithHostname",
    "JarmHash",
    "ListeningSocket",
    "ListeningSocketUsesPort",
    "ServiceIdentifiedAsPlatform",
    "MACAddress",
    "MACAddressAssignedToHost",
    "MitreAttackTactic",
    "MitreTacticIncludesTechnique",
    "MitreAttackTechnique",
    "MitreSubtechniqueOf",
    "MitreAttackCampaign",
    "MitreCampaignUsesSoftware",
    "MitreCampaignUsesTechnique",
    "MitreAttackSoftware",
    "MitreSoftwareUsesTechnique",
    "MitreCampaignAttributedTo",
    "NetworkService",
    "NetworkServiceRunsOnPort",
    "OpenPortHasJarmHash",
    "OpenPortPresentsBanner",
    "OpenPortPresentsX509Certificate",
    "Organisation",
    "OrganisationAssignedCVSSToCVE",
    "OrganisationPublishedThreatReport",
    "OrganisationReportedExploitationOfCVE",
    "Port",
    "Report",
    "ReportMentionsCountry",
    "ReportMentionsCVE",
    "ReportMentionsIntrusionSet",
    "ReportMentionsIP",
    "ReportMentionsSector",
    "ReportMentionsTechnique",
    "ReportIdentifiesIntrusionSet",
    "ReportIdentifiesTechnique",
    "ReportIdentifiesCVE",
    "ReportIdentifiesVictimSector",
    "ReportIdentifiesVictimCountry",
    "ReportIdentifiesSponsorCountry",
    "ReportIdentifiesCyberHarm",
    "ReportIdentifiesCampaign",
    "ReportIdentifiesSoftware",
    "Sector",
    "ServiceHostsURL",
    "ThreatActor",
    "ThreatActorAttributedToNation",
    "ThreatActorIsOfType",
    "X509Certificate",
    "URLNode",
    "UrlRedirectsToUrl",
    "UserAccount",
    "UserAccountAuthorizedOnHost",
]



from typing import ClassVar, Optional
from warnings import warn

import pandas as pd
from neontology import BaseNode

from .dataorigin import DataOrigin, OriginGenerated


class OntolocyNode(BaseNode, validate_default=True):
    __primarylabel__: ClassVar[Optional[str]] = None

    @classmethod
    def ingest_df(
        cls,
        df: pd.DataFrame,
        data_origin: Optional[DataOrigin] = None,
        deduplicate=True,
    ) -> list:
        nodes = cls.merge_df(df, deduplicate=deduplicate)

        if data_origin is not None:
            data_origin.merge()

            rels = [OriginGenerated(source=data_origin, target=x) for x in nodes]

            OriginGenerated.merge_relationships(rels, target_type=cls)

        return nodes

    def ingest(
        self,
        data_origin: Optional[DataOrigin] = None,
    ) -> None:
        # first merge in this node
        self.merge()

        # if a data_origin is specified, merge that and then create the OriginGenerated relationship
        if data_origin is not None:
            data_origin.merge()

            rel = OriginGenerated(source=data_origin, target=self)

            rel.merge()

    def get_identifier(self) -> str:
        warn("'get_identifier' method is deprecated", DeprecationWarning)
        return str(self.get_primary_property_value())

    def __str__(self) -> str:
        return str(self.get_primary_property_value())



from datetime import datetime
from typing import ClassVar, Optional
from uuid import UUID

from pydantic import HttpUrl, ValidationInfo, field_validator

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from ..utils import generate_deterministic_uuid
from .cve import CVE


class Exploit(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "Exploit"

    name: str
    url: HttpUrl

    published_date: Optional[datetime] = None

    unique_id: Optional[UUID] = None

    @field_validator("unique_id")
    def generate_unique_id(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data

        if v is None:
            key_values = [
                values["name"],
                values["url"],
            ]

            v = generate_deterministic_uuid(key_values)

        return v

    def __str__(self) -> str:
        return self.name


#
# OUTGOING RELATIONSHIPS
#


class ExploitExploitsVulnerability(OntolocyRelationship):
    source: Exploit
    target: CVE

    __relationshiptype__: ClassVar[str] = "EXPLOIT_EXPLOITS_VULNERABILITY"



from typing import ClassVar, List, Optional

from ..node import OntolocyNode

sectors = {
    "agriculture": {
        "sector_id": "agriculture",
        "name": "Agriculture",
        "keywords": ["agriculture", "farming"],
    },
    "aerospace": {
        "sector_id": "aerospace",
        "name": "Aerospace",
        "keywords": ["aerospace", "airport"],
    },
    "automotive": {
        "sector_id": "automotive",
        "name": "Automotive",
        "keywords": ["automotive"],
    },
    "chemical": {
        "sector_id": "chemical",
        "name": "Chemical",
        "keywords": ["chemical"],
    },
    "commercial": {
        "sector_id": "commercial",
        "name": "Commercial",
        "keywords": ["commercial"],
    },
    "communications": {
        "sector_id": "communications",
        "name": "Communications",
        "keywords": [],
    },
    "construction": {
        "sector_id": "construction",
        "name": "Construction",
        "keywords": ["construction"],
    },
    "defense": {
        "sector_id": "defense",
        "name": "Defense",
        "keywords": ["defense industrial base"],
    },
    "education": {
        "sector_id": "education",
        "name": "Education",
        "keywords": ["education", "schoool", "university", "universities"],
    },
    "energy": {
        "sector_id": "energy",
        "name": "Energy",
        "keywords": ["energy"],
    },
    "entertainment": {
        "sector_id": "entertainment",
        "name": "Entertainment",
        "keywords": ["entertainment"],
    },
    "financial-services": {
        "sector_id": "financial-services",
        "name": "Financial Services",
        "keywords": ["financial services"],
    },
    "emergency-services": {
        "sector_id": "emergency-services",
        "name": "Emergency Services",
        "keywords": ["emergency services", "ambulance"],
    },
    "government-local": {
        "sector_id": "government-local",
        "name": "Local Government",
        "keywords": ["local government"],
    },
    "government-national": {
        "sector_id": "government-national",
        "name": "National Government",
        "keywords": ["national government"],
    },
    "government-public-services": {
        "sector_id": "government-public-services",
        "name": "Public Services",
        "keywords": ["public services"],
    },
    "government-regional": {
        "sector_id": "government-regional",
        "name": "Regional Govenment",
        "keywords": ["government regional"],
    },
    "healthcare": {
        "sector_id": "healthcare",
        "name": "Healthcare",
        "keywords": ["healthcare", "hospital", "care home"],
    },
    "hospitality-leisure": {
        "sector_id": "hospitality-leisure",
        "name": "Leisure and Hospitality",
        "keywords": ["leisure", "hospitality", "restaurant", "cafe"],
    },
    "dams": {
        "sector_id": "dams",
        "name": "Dams",
        "keywords": ["dam"],
    },
    "nuclear": {
        "sector_id": "nuclear",
        "name": "Nuclear",
        "keywords": ["nuclear"],
    },
    "water": {
        "sector_id": "water",
        "name": "Water",
        "keywords": ["water"],
    },
    "insurance": {
        "sector_id": "insurance",
        "name": "Insurance",
        "keywords": ["insurance"],
    },
    "manufacturing": {
        "sector_id": "manufacturing",
        "name": "Manufacturing",
        "keywords": ["manufacturing"],
    },
    "mining": {
        "sector_id": "mining",
        "name": "Mining",
        "keywords": ["mining"],
    },
    "non-profit": {
        "sector_id": "non-profit",
        "name": "Non Profit",
        "keywords": ["non profit", "not for profit", "charity"],
    },
    "pharmaceuticals": {
        "sector_id": "pharmaceuticals",
        "name": "Pharmaceuticals",
        "keywords": ["pharmaceuticals"],
    },
    "retail": {
        "sector_id": "retail",
        "name": "Retail",
        "keywords": ["retail"],
    },
    "technology": {
        "sector_id": "technology",
        "name": "Technology",
        "keywords": ["technology"],
    },
    "telecommunications": {
        "sector_id": "telecommunications",
        "name": "Telecommunications",
        "keywords": ["telecommunications"],
    },
    "transportation": {
        "sector_id": "transportation",
        "name": "Transportation",
        "keywords": ["transportation", "rail"],
    },
    "utilities": {
        "sector_id": "utilities",
        "name": "Utilities",
        "keywords": ["utilities"],
    },
}


class Sector(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "Sector"

    name: str
    unique_id: str
    keywords: Optional[List[str]]



from typing import ClassVar, Optional

from ..node import OntolocyNode


class CWE(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "cwe_id"
    __primarylabel__: ClassVar[Optional[str]] = "CWE"

    cwe_id: int
    description: str
    name: str
    abstraction: str
    structure: str
    status: str



from enum import Enum
from typing import ClassVar, Optional

from pydantic import ValidationInfo, field_validator

from ..node import OntolocyNode

actor_type_taxonomy = {
    "nation-state": {
        "name": "Nation State",
        "description": "An actor backed by a nation state, to achieve aims aligned with state objectives.",
    },
    "criminal-group": {
        "name": "Criminal Group",
        "description": "A group of criminals working together to conduct illegal cyber activities.",
    },
    "hacktivists": {
        "name": "Hacktivists",
        "description": "An individual or group who's cyber activities are primarily ideologically driven.",
    },
    "commercial-provider": {
        "name": "Commercial Provider",
        "description": (
            "An actor who sells offensive cyber services to others,"
            " and those services are deemed legal in the jurisdiction that they operate within."
            " Also known as 'hackers-for-hire'"
        ),
    },
}


class ActorTypeEnum(str, Enum):
    nation_state = "nation-state"
    criminal_group = "criminal-group"
    hacktivists = "hacktivists"
    commercial_provider = "commercial-provider"


class ActorType(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "ActorType"

    unique_id: ActorTypeEnum
    name: Optional[str] = None
    description: Optional[str] = None

    def __str__(self) -> str:
        if self.name is not None:
            return self.name
        else:
            return self.unique_id

    @field_validator("name")
    def set_actor_name(cls, v, info: ValidationInfo):
        values = info.data
        if v is None and "unique_id" in values:
            return actor_type_taxonomy[values["unique_id"]]["name"]
        else:
            return v

    @field_validator("description")
    def set_actor_description(cls, v, info: ValidationInfo):
        values = info.data
        if v is None and "unique_id" in values:
            return actor_type_taxonomy[values["unique_id"]]["description"]
        else:
            return v



from datetime import datetime
from typing import ClassVar, Optional

from pydantic import HttpUrl, field_validator

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .mitreattacktechnique import MitreAttackTechnique


class MitreAttackSoftware(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "stix_id"
    __primarylabel__: ClassVar[str] = "MitreAttackSoftware"

    stix_id: str
    stix_type: str
    stix_created: datetime
    stix_modified: datetime
    stix_spec_version: str = "2.1"
    stix_revoked: Optional[bool] = False

    attack_spec_version: str
    attack_version: str
    attack_id: str

    ref_url: HttpUrl
    name: str
    description: Optional[str] = None

    @field_validator("stix_revoked")
    @classmethod
    def set_false(cls, v):
        if v is None:
            return False
        else:
            return v


#
# OUTGOING RELATIONSHIPS
#


class MitreSoftwareUsesTechnique(OntolocyRelationship):
    source: MitreAttackSoftware
    target: MitreAttackTechnique

    __relationshiptype__: ClassVar[str] = "MITRE_SOFTWARE_USES_TECHNIQUE"



from enum import Enum
from ipaddress import ip_address
from typing import ClassVar, Optional
from uuid import UUID, uuid4

from pydantic import IPvAnyAddress, StringConstraints, ValidationInfo, field_validator
from typing_extensions import Annotated

from ontolocy.models.port import Port
from ontolocy.node import OntolocyNode
from ontolocy.relationship import OntolocyRelationship
from ontolocy.utils import generate_deterministic_uuid

from .banner import Banner
from .cpe import CPE
from .jarmhash import JarmHash
from .url import URLNode
from .x509certificate import X509Certificate


class ListeningSocketProtocolEnum(str, Enum):
    tcp = "tcp"
    udp = "udp"
    sctp = "sctp"


class ListeningSocket(OntolocyNode):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "ListeningSocket"

    protocol: ListeningSocketProtocolEnum
    port_number: int
    ip_address: IPvAnyAddress
    private: Optional[bool] = None
    namespace: Optional[str] = None
    ip_address_unique_id: Optional[
        UUID
    ] = None  # for private IPs, uniquely identify the IP to avoid collisions

    unique_id: Optional[UUID] = None

    def __str__(self) -> str:
        return f"{self.ip_address}:{self.port_number} ({self.protocol.value})"

    @field_validator("unique_id")
    def generate_socket_uuid(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data
        if v is None:
            key_values = [
                values["protocol"],
                values["port_number"],
                values["ip_address"],
                values["ip_address_unique_id"],
            ]

            v = generate_deterministic_uuid(key_values)

        return v

    @field_validator("private")
    def mark_private(cls, v, info: ValidationInfo):
        values = info.data
        if v is None and "ip_address" in values:
            return ip_address(values["ip_address"]).is_private
        else:
            return v

    @field_validator("namespace")
    def set_namespace(cls, v, info: ValidationInfo):
        values = info.data

        if values["private"] is False:
            return None

        elif v is None:
            return str(uuid4())

        else:
            return v

    @field_validator("ip_address_unique_id")
    def generate_ip_id(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data
        if v is None:
            key_values = [values["ip_address"], values["namespace"]]

            v = generate_deterministic_uuid(key_values)

        return v


#
# OUTGOING RELATIONSHIPS
#


class ListeningSocketUsesPort(OntolocyRelationship):
    source: ListeningSocket
    target: Port

    __relationshiptype__: ClassVar[str] = "LISTENING_SOCKET_USES_PORT"


class OpenPortPresentsBanner(OntolocyRelationship):
    source: ListeningSocket
    target: Banner

    __relationshiptype__: ClassVar[str] = "OPEN_PORT_PRESENTS_BANNER"


class OpenPortPresentsX509Certificate(OntolocyRelationship):
    source: ListeningSocket
    target: X509Certificate

    __relationshiptype__: ClassVar[str] = "OPEN_PORT_PRESENTS_X509_CERTIFICATE"


class OpenPortHasJarmHash(OntolocyRelationship):
    source: ListeningSocket
    target: JarmHash

    __relationshiptype__: ClassVar[str] = "OPEN_PORT_HAS_JARM_HASH"


class ServiceHostsURL(OntolocyRelationship):
    source: ListeningSocket
    target: URLNode
    status_code: Optional[int] = None

    __relationshiptype__: ClassVar[str] = "SERVICE_HOSTS_URL"


class ServiceIdentifiedAsPlatform(OntolocyRelationship):
    source: ListeningSocket
    target: CPE
    status_code: Optional[int] = None

    cpe: Optional[
        Annotated[
            str,
            StringConstraints(
                pattern=(
                    r"(cpe:2\.3:[aho\*\-](:(((\?*|\*?)([a-zA-Z0-9\-\._]|"  # noqa: F722
                    r"(\\[\\\*\?!#$$%&'\(\)\+,/:;<=>@\[\]\^`\{\|}~]))+(\?*|\*?))|[\*\-])){5}"
                    r"(:(([a-zA-Z]{2,3}(-([a-zA-Z]{2}|[0-9]{3}))?)|[\*\-]))(:(((\?*|\*?)"
                    r"([a-zA-Z0-9\-\._]|(\\[\\\*\?!#$$%&'\(\)\+,/:;<=>@\[\]\^`\{\|}~]))+(\?*|\*?))|[\*\-])){4})"
                )
            ),
        ]
    ]

    @field_validator("cpe", mode="before")
    def set_cpe(cls, v):
        # hack to handle CPEs with colons in (which will be escaped with a backslash)
        v = v.replace(r"\:", r"\;")

        cpe_parts = v.split(":")

        if cpe_parts[0] != "cpe":
            raise ValueError("Doesn't look like a valid CPE")

        if cpe_parts[1] == "2.3":
            v = "cpe"

            for idx in range(1, 13):
                v += ":"
                if idx < len(cpe_parts):
                    v += cpe_parts[idx]
                else:
                    v += "*"

        else:
            raise ValueError("Doesn't look like a compatible CPE 2.3 format")

        return v

    __relationshiptype__: ClassVar[str] = "SERVICE_IDENTIFIED_AS_PLATFORM"



from enum import Enum
from typing import ClassVar, Optional
from uuid import UUID

from pydantic import ValidationInfo, field_validator

from ..node import OntolocyNode
from ..utils import generate_deterministic_uuid


class PortProtocolEnum(str, Enum):
    tcp = "tcp"
    udp = "udp"
    sctp = "sctp"


class Port(OntolocyNode):
    __primarylabel__: ClassVar[str] = "Port"
    __primaryproperty__: ClassVar[str] = "unique_id"

    port_number: int
    protocol: PortProtocolEnum

    unique_id: Optional[UUID] = None

    def __str__(self) -> str:
        return f"{self.port_number} ({self.protocol.value})"

    @field_validator("unique_id")
    def generate_socket_uuid(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data

        if v is None:
            key_values = [
                values["port_number"],
                values["protocol"],
            ]

            v = generate_deterministic_uuid(key_values)

        return v



from typing import ClassVar, Optional
from uuid import UUID

from neontology import BaseNode, BaseRelationship
from pydantic import ValidationInfo, field_validator

from .utils import generate_deterministic_uuid


class DataOrigin(BaseNode, validate_default=True):
    __primaryproperty__: ClassVar[str] = "unique_id"
    __primarylabel__: ClassVar[Optional[str]] = "DataOrigin"

    name: str
    reference: Optional[str] = None
    license: Optional[str] = None
    sharing: Optional[str] = None

    unique_id: Optional[UUID] = None

    @field_validator("unique_id")
    def generate_unique_id(cls, v: Optional[UUID], info: ValidationInfo) -> UUID:
        values = info.data
        if v is None:
            key_values = [
                values["name"],
                values["reference"],
                values["license"],
                values["sharing"],
            ]

            v = generate_deterministic_uuid(key_values)

        return v


class OriginGenerated(BaseRelationship):
    __relationshiptype__: ClassVar[Optional[str]] = "ORIGIN_GENERATED"

    source: DataOrigin
    target: BaseNode



from typing import ClassVar, Optional

from ..node import OntolocyNode
from ..relationship import OntolocyRelationship
from .cwe import CWE
from .mitreattacktechnique import MitreAttackTechnique


class CAPECPattern(OntolocyNode):

    __primaryproperty__: ClassVar[str] = "capec_id"
    __primarylabel__: ClassVar[Optional[str]] = "CAPECPattern"

    capec_id: int
    description: Optional[str] = None
    name: str
    likelihood_of_attack: Optional[str] = None
    typical_severity: Optional[str] = None


#
# OUTGOING RELATIONSHIPS
#


class CAPECPatternMapsToAttackTechnique(OntolocyRelationship):
    source: CAPECPattern
    target: MitreAttackTechnique

    __relationshiptype__: ClassVar[str] = "CAPEC_PATTERN_MAPS_TO_ATTACK_TECHNIQUE"


class CAPECPatternRelatesToCWE(OntolocyRelationship):
    source: CAPECPattern
    target: CWE

    __relationshiptype__: ClassVar[str] = "CAPEC_PATTERN_RELATES_TO_CWE"



from typing import Optional
from uuid import UUID, uuid5

import pandas as pd
from neontology import init_neontology


def init_ontolocy(
    neo4j_uri: Optional[str] = None,
    neo4j_username: Optional[str] = None,
    neo4j_password: Optional[str] = None,
) -> None:
    """Initialise ontolocy.

    Wraps the init_neontology function.

    Args:
        neo4j_uri (Optional[str], optional): Neo4j URI to connect to. Defaults to None.
        neo4j_username (Optional[str], optional): Neo4j username. Defaults to None.
        neo4j_password (Optional[str], optional): Neo4j password. Defaults to None.
    """

    init_neontology(neo4j_uri, neo4j_username, neo4j_password)


def generate_deterministic_uuid(values: list) -> UUID:

    # we use an arbitrary 'ontolocy' specific namespace
    namespace = UUID("8e43adb2-a389-4e4b-8012-3bc5702fb832")

    # create a single long string from all the values provided
    value_string = "".join([str(x) for x in values])

    generated_id = uuid5(namespace, value_string)

    return generated_id


def explode_map_dfs(df, to_map):

    input_df = df.copy()

    to_map_df = pd.concat(to_map)

    # keep a record of the original order
    input_df.insert(0, "ontolocy_merging_order", range(0, len(input_df)))

    merge_cols = list(set(to_map_df.columns) & set(input_df.columns))

    output_df = (
        input_df.merge(
            to_map_df,
            how="inner",
            on=merge_cols,
        )
        .sort_values("ontolocy_merging_order", ignore_index=True)
        .copy()
    ).drop(columns="ontolocy_merging_order")

    return output_df
